import threading

try:
    import queue
except ImportError:
    import Queue as queue


class Exit(Exception):
    pass


class ExitableQueue(object):
    def __init__(self, exit, maxsize=0, poll=0.1):
        self._q = queue.Queue(maxsize)
        self._exit = exit
        self.poll = poll

    def put(self, item):
        while True:
            try:
                return self._q.put(item, timeout=self.poll)
            except queue.Full:
                if self._exit.is_set():
                    raise Exit

    def get(self):
        while True:
            try:
                return self._q.get(timeout=self.poll)
            except queue.Empty:
                if self._exit.is_set():
                    raise Exit

    def put_nowait(self):
        return self._q.put(block=False)


class ExitableBoundedSemaphore(object):
    def __init__(self, exit, value=1, poll=0.1):
        self._sem = threading.BoundedSemaphore(value)
        self._exit = exit
        self.poll = poll

    def acquire(self):
        while True:
            if self._sem.acquire():
                return

            elif self._exit.is_set():
                raise Exit

    def release(self):
        return self._sem.release()



import threading
from abc import ABCMeta, abstractmethod

from six import with_metaclass
from loky import ProcessPoolExecutor

from . import exitable


class Task(with_metaclass(ABCMeta, object)):
    @abstractmethod
    def __next__(self):
        raise NotImplementedError

    def on_task_start(self):
        pass

    def on_task_end(self):
        pass

    def on_job_start(self, job):
        pass

    def on_job_end(self, job, v):
        pass

    def on_job_error(self, job, e):
        pass

    def next_task(self):
        return None

    timeout = None


class SingleTask(Task):
    @abstractmethod
    def job(self):
        raise NotImplementedError

    def __next__(self):
        if getattr(self, "_already_executed", False):
            raise StopIteration

        self._already_executed = True
        return self.job()


class TaskWrapper(object):
    def __init__(self, task):
        self.raw = task
        self.job_count = 0
        self.lock = threading.Lock()

    def incr(self):
        with self.lock:
            self.job_count += 1

    def decr(self, on_end):
        with self.lock:
            self.job_count -= 1
            if self.job_count < 0:
                on_end()


class Counter(object):
    def __init__(self, value):
        self.lock = threading.Lock()
        self.ev = threading.Event()
        self.value = value
        self.check_value()

    def check_value(self):
        if self.value <= 0:
            self.ev.set()
        else:
            self.ev.clear()

    def incr(self):
        with self.lock:
            self.value += 1
            self.check_value()

    def decr(self):
        with self.lock:
            self.value -= 1
            self.check_value()

    def join(self):
        return self.ev.wait()


class Daemon(threading.Thread):
    def run(self):
        try:
            while True:
                self._main()

        except exitable.Exit:
            pass


class MoveThread(Daemon):
    def __init__(self, q):
        super(MoveThread, self).__init__()
        self.q = q

    def _main(self):
        self.q._sem.acquire()
        task = self.q._pendings.get()
        self.q._workings.put(task)

        self.q._ioloop.add_callback(task.raw.on_task_start)


class OnTaskEnd(object):
    def __init__(self, q, task):
        self.q = q
        self.task = task

    def task_end_callback(self):
        self.task.raw.on_task_end()

        next_task = self.task.raw.next_task()
        if next_task is not None:
            self.q.put(next_task)

        self.q._sem.release()
        self.q._cnt.decr()

    def __call__(self):
        self.q._ioloop.add_callback(self.task_end_callback)


class WorkerThread(Daemon):
    def __init__(self, q):
        super(WorkerThread, self).__init__()
        self.q = q

    def task_end(self, task):
        return OnTaskEnd(self.q, task)

    def _main(self):
        task = self.q._workings.get()
        try:
            job = task.raw.__next__()
        except StopIteration:
            task.decr(self.task_end(task))
            return

        task.incr()
        self.q._workings.put(task)
        fut = self.q._pool.submit(job)
        self.q._ioloop.add_callback(task.raw.on_job_start, job)

        try:
            result = fut.result(timeout=task.raw.timeout)
            self.q._ioloop.add_callback(task.raw.on_job_end, job, result)
        except Exception as e:
            self.q._ioloop.add_callback(task.raw.on_job_error, job, e)

        task.decr(self.task_end(task))


class TaskQueue(object):
    def __init__(self, workers, ioloop):
        self._exit = threading.Event()
        self._pendings = exitable.ExitableQueue(self._exit)
        self._workings = exitable.ExitableQueue(self._exit, workers)
        self._sem = exitable.ExitableBoundedSemaphore(self._exit, workers)
        self._pool = ProcessPoolExecutor(workers)
        self._cnt = Counter(0)
        self._ioloop = ioloop

        self._workers = [MoveThread(self)]
        self._workers += [WorkerThread(self) for _ in range(workers)]

    def put(self, task):
        self._cnt.incr()
        self._pendings.put(TaskWrapper(task))

    def __enter__(self):
        for w in self._workers:
            w.start()

        return self

    def __exit__(self, *args, **kwargs):
        self._exit.set()
        for w in self._workers:
            w.join()

        self._pool.shutdown()

    def join(self):
        return self._cnt.join()


class TestJob(object):
    def __init__(self, name, i):
        self.name = name
        self.i = i

    def __call__(self):
        return self.name, self.i


class TestTask(Task):
    def __init__(self, name, count=5):
        self.count = count
        self.i = 0
        self.name = name

    def __next__(self):
        if self.i >= self.count:
            raise StopIteration

        self.i += 1
        return TestJob(self.name, self.i)

    def on_task_start(self):
        print("task start")  # noqa: T003

    def on_task_end(self):
        print("task end")  # noqa: T003

    def on_job_start(self, job):
        print("job start")  # noqa: T003

    def on_job_end(self, job, v):
        print("job end", v)  # noqa: T003

    def on_job_error(self, job, e):
        print("job error", e)  # noqa: T003


def main():
    with TaskQueue(20) as q:
        for i in range(100):
            q.put(TestTask(i))

        q.join()


if __name__ == "__main__":
    main()



import os
import re
import time
from io import BytesIO
from cgi import parse_header
from tempfile import NamedTemporaryFile

from rdkit import Chem
from tornado import gen, web
from rdkit.Chem import Draw
from rdkit.Chem.rdDistGeom import EmbedMolecule
from rdkit.Chem.rdForceFieldHelpers import UFFOptimizeMolecule, MMFFOptimizeMolecule

from ..db import Phase, transaction, issue_text_id
from .common import SSEHandler, RequestHandler
from ..task_queue import Task, SingleTask

MEGA = 1024 * 1024
SMI_FIELDS = re.compile(br"^(\S+)\s+(.+)?")


def read_smiles(bs):
    for i, line in enumerate(BytesIO(bs), 1):
        fields = SMI_FIELDS.match(line)
        if fields is None:
            yield ValueError("parse failed on line {}".format(i)), None
            continue

        smi, name = fields.groups()
        if name is None:
            name = smi

        mol = Chem.MolFromSmiles(smi)
        if mol is None:
            yield ValueError(
                "SMILES parse failed on line {}: {}".format(
                    i, line.decode("UTF-8").strip()), ), None
            continue
        yield mol, name.decode("UTF-8")


def read_sdf(bs):
    with NamedTemporaryFile() as tmp:
        tmp.write(bs)
        tmp.flush()

        for i, mol in enumerate(
                Chem.SDMolSupplier(tmp.name, removeHs=False), 1):
            if mol is None:
                yield ValueError(
                    "SDF parser failed on {}-th molecule".format(i)), None
                continue

            if mol.HasProp("_Name"):
                name = mol.GetProp("_Name")
            else:
                name = Chem.MolToSmiles(mol)

            yield mol, name


class ParseTask(SingleTask):
    def __init__(self, text_id, filename, body, gen3D, desalt, conn, reader,
                 parse_timeout, prepare_timeout, molecule_limit):

        self.conn = conn
        self.text_id = text_id
        self.filename = filename
        self.body = body
        self.gen3D = gen3D
        self.desalt = desalt
        self.reader = reader
        self.timeout = parse_timeout
        self.prepare_timeout = prepare_timeout
        self.molecule_limit = molecule_limit

    def insert_file(self):
        is3D = self.reader != read_smiles or self.gen3D

        with transaction(self.conn) as cur:
            cur.execute("""
            INSERT INTO file (text_id, name, created_at, gen3D, is3D, desalt, phase)
            VALUES (?, ?, ?, ?, ?, ?, ?)
            """, (self.text_id, self.filename, int(time.time()), self.gen3D,
                  is3D, self.desalt, Phase.PENDING.value))
            self.file_id = cur.lastrowid

    def on_task_start(self):
        with transaction(self.conn) as cur:
            cur.execute(
                "UPDATE file SET phase = ? WHERE id = ?",
                (Phase.IN_PROGRESS.value, self.file_id), )

    def on_job_end(self, job, v):
        self.mols, errors = v
        with transaction(self.conn) as cur:
            cur.execute(
                "UPDATE file SET total = ? WHERE id = ?",
                (len(self.mols), self.file_id), )
            for err in errors:
                cur.execute("""
                INSERT INTO file_error (file_id, error)
                VALUES (?, ?)
                """, (self.file_id, str(err)))

    def on_job_error(self, job, e):
        se = str(e)
        if len(se) == 0:
            se = repr(e)

        with transaction(self.conn) as cur:
            cur.execute(
                "UPDATE file SET phase = ?, total = 0 WHERE id = ?",
                (Phase.ERROR.value, self.file_id, ), )

            cur.execute(
                "INSERT INTO file_error (file_id, error) VALUES (?, ?)",
                (self.file_id, "parse error: {}".format(se)), )

    def next_task(self):
        if not hasattr(self, "mols"):
            return

        return PrepareTask(
            file_id=self.file_id,
            mols=self.mols,
            gen3D=self.gen3D,
            desalt=self.desalt,
            conn=self.conn,
            timeout=self.prepare_timeout, )

    def job(self):
        return ParseJob(
            body=self.body,
            reader=self.reader,
            molecule_limit=self.molecule_limit, )


class ParseJob(object):
    def __init__(self, body, reader, molecule_limit):
        self.body = body
        self.reader = reader
        self.molecule_limit = molecule_limit

    def __call__(self):
        mols, errors = [], []
        nth = 0
        for i, (mol, name) in enumerate(self.reader(self.body)):
            if self.molecule_limit is not None and i >= self.molecule_limit:
                errors.append(
                    "number of molecule limit: using first {} molecules".
                    format(
                        self.molecule_limit, ))
                break

            if not isinstance(mol, Chem.Mol):
                errors.append(mol)
                continue

            mols.append((mol, nth, name.strip()))
            nth += 1

        return mols, errors


class PrepareTask(Task):
    def __init__(self, file_id, mols, gen3D, desalt, conn, timeout):
        self.file_id = file_id
        self.mols = mols
        self.gen3D = gen3D
        self.desalt = desalt
        self.conn = conn
        self.timeout = timeout

    def on_job_end(self, job, result):
        uff, mol, nth, name = result
        if self.gen3D:
            ff = "UFF" if uff else "MMFF"
        else:
            ff = None
        with transaction(self.conn) as cur:
            cur.execute("""
            INSERT INTO molecule (file_id, nth, forcefield, name, mol)
            VALUES (?, ?, ?, ?, ?)
            """, (self.file_id, nth, ff, name, mol))

    def on_job_error(self, job, err):
        se = str(err)
        if len(se) == 0:
            se = repr(err)

        with transaction(self.conn) as cur:
            cur.execute(
                "UPDATE file SET total = total - 1 WHERE id = ?",
                (self.file_id, ), )
            cur.execute(
                "INSERT INTO file_error (file_id, error) VALUES (?, ?)",
                (self.file_id, "{}: prepare: {}".format(job.name, se)), )

    def on_task_end(self):
        with transaction(self.conn) as cur:
            cur.execute(
                "UPDATE file SET phase = ? WHERE id = ?",
                (Phase.DONE.value, self.file_id), )

    def __next__(self):
        if len(self.mols) == 0:
            raise StopIteration

        (mol, nth, name), self.mols = self.mols[0], self.mols[1:]
        return PrepareJob(mol, nth, name, self.gen3D, self.desalt)


def desalt(mol):
    mols = Chem.GetMolFrags(mol, asMols=True)
    return max(mols, key=lambda m: m.GetNumAtoms())


def optimize(mol, f):
    for _ in range(10):
        r = f(mol, maxIters=1000)
        if r == 0:
            return
        elif r != 1:
            raise ValueError("{} failed".format(f.__name__))

    raise ValueError("{} optimize not converged".format(f.__name__))


def gen3D(mol):
    mol = Chem.AddHs(mol)
    if EmbedMolecule(mol) != 0:
        raise ValueError("EmbedMolecule failed")

    uff = False

    try:
        optimize(mol, MMFFOptimizeMolecule)
    except ValueError:
        uff = True

    if uff:
        optimize(mol, UFFOptimizeMolecule)

    return uff, mol


class PrepareJob(object):
    def __init__(self, mol, nth, name, gen3D, desalt):
        self.mol = mol
        self.nth = nth
        self.name = name
        self.gen3D = gen3D
        self.desalt = desalt

    def __call__(self):
        mol = self.mol
        if self.desalt:
            mol = desalt(mol)

        if self.gen3D:
            uff, mol = gen3D(mol)
            return uff, mol, self.nth, self.name
        else:
            return None, mol, self.nth, self.name


class FileHandler(RequestHandler):
    SMI_EXT = {".smi", ".smiles"}
    SDF_EXT = {".sdf", ".sd", ".mol"}

    def post(self):
        gen3D = self.get_flag("gen3D", False)
        desalt = self.get_flag("desalt", True)

        f = self.request.files.get("file")
        if f is None:
            self.fail(400, "no file parameter")

        f = f[0]
        limit_mb = self.application.file_size_limit
        limit_b = limit_mb * MEGA
        if len(f.body) > limit_b:
            self.fail(400, "file size too large (> {}MB)".format(limit_mb))

        ext = os.path.splitext(f.filename)[-1].lower()

        if ext in self.SMI_EXT:
            reader = read_smiles
        elif ext in self.SDF_EXT:
            reader = read_sdf
        else:
            self.fail(400, "unknown extension: {}".format(ext))

        text_id = issue_text_id()
        task = ParseTask(
            text_id=text_id,
            filename=f.filename,
            body=f.body,
            gen3D=gen3D,
            desalt=desalt,
            conn=self.db,
            reader=reader,
            parse_timeout=self.application.parse_timeout,
            prepare_timeout=self.application.prepare_timeout,
            molecule_limit=self.application.molecule_limit, )
        task.insert_file()
        self.put(task)

        self.json(id=text_id)


class FileIdHandler(SSEHandler):
    def get(self, id):
        with self.transaction() as cur:
            cur.execute("SELECT id, name FROM file WHERE text_id = ? LIMIT 1",
                        (id, ))
            result = cur.fetchone()

        if result is None:
            self.fail(404, "no id")

        self.file_id, self.filename = result

        accept, _ = parse_header(self.request.headers["Accept"])

        if accept == "text/event-stream":
            return self.get_sse(id)

        else:
            return self.get_json(id)

    @gen.coroutine
    def get_sse(self, id):
        self.init_sse()
        while True:
            with self.transaction() as cur:
                cur.execute("""
                SELECT total, phase, count(molecule.file_id)
                FROM file LEFT OUTER JOIN molecule ON file.id = molecule.file_id
                WHERE file.id = ?
                LIMIT 1
                """, (self.file_id, ))

                total, phase, current = cur.fetchone()

            yield self.publish(
                total=total, name=self.filename, phase=phase, current=current)
            if phase == Phase.ERROR.value or phase == Phase.DONE.value:
                raise web.Finish
            yield gen.sleep(0.5)

    def get_json(self, id):
        with self.transaction() as cur:
            cur.execute(
                "SELECT name, gen3D, is3D, desalt, phase FROM file WHERE id = ? LIMIT 1",
                (self.file_id, ), )
            name, gen3D, is3D, desalt, phase = cur.fetchone()

            cur.execute(
                "SELECT name, forcefield FROM molecule WHERE file_id = ? ORDER BY nth",
                (self.file_id, ), )
            mols = [{"name": n, "forcefield": f} for n, f in cur.fetchall()]

            cur.execute(
                "SELECT error FROM file_error WHERE file_id = ? ORDER BY id",
                (self.file_id, ), )
            errors = [e for e, in cur.fetchall()]

        self.json(
            name=name,
            gen3D=bool(gen3D),
            desalt=bool(desalt),
            is3D=bool(is3D),
            mols=mols,
            errors=errors,
            phase=phase, )


class FileIdExtHandler(RequestHandler):
    EXTS = {"sdf", "smi"}

    def get(self, text_id, ext):
        ext = ext.lower()
        if ext not in self.EXTS:
            self.fail(400, "unknown extension")

        with self.transaction() as cur:
            cur.execute("SELECT id FROM FILE WHERE text_id = ? LIMIT 1",
                        (text_id, ))
            result = cur.fetchone()
            if result is None:
                self.fail(404, "not found")

            file_id, = result

            cur.execute("""
            SELECT name, mol, forcefield
            FROM molecule
            WHERE file_id = ?
            ORDER BY nth
            """, (file_id, ))

            if ext == "sdf":
                self.get_sdf(cur)
            elif ext == "smi":
                self.get_smi(cur)
            else:
                self.fail(500, "BUG: unknown extension: {}".format(ext))

    def get_sdf(self, cur):
        self.set_header("content-type", "chemical/x-mdl-sdfile")
        with NamedTemporaryFile() as temp:
            writer = Chem.SDWriter(temp.name)
            while True:
                result = cur.fetchone()
                if result is None:
                    break

                name, mol, forcefield = result
                mol.SetProp("_Name", name)
                if forcefield:
                    mol.SetProp("ForceField", forcefield)

                writer.write(mol)

            writer.close()
            temp.seek(0)
            self.write(temp.read())

    def get_smi(self, cur):
        self.set_header("content-type", "chemical/x-daylight-smiles")
        while True:
            result = cur.fetchone()
            if result is None:
                break

            name, mol, _ = result
            self.write("{} {}\n".format(Chem.MolToSmiles(Chem.RemoveHs(mol)), name))


class FileIdNthExtHandler(RequestHandler):
    EXTS = {"png", "mol"}

    def get(self, id, nth, ext):
        ext = ext.lower()
        if ext not in self.EXTS:
            self.fail(400, "unknown extension")

        nth = int(nth)

        with self.transaction() as cur:
            cur.execute("""
            SELECT name, mol, forcefield
            FROM molecule
            WHERE nth = ?
            AND file_id = (SELECT id FROM FILE WHERE text_id = ?)
            LIMIT 1""", (nth, id))  # noqa: Q445

            result = cur.fetchone()

        if result is None:
            self.fail(404, "not found")

        self.name, self.mol, self.forcefield = result

        if ext == "png":
            self.get_png()
        elif ext == "mol":
            self.get_mol()
        else:
            self.fail(500, "BUG: unknown extension: {}".format(ext))

    def get_png(self):
        self.mol.RemoveAllConformers()
        img = Draw.MolToImage(self.mol, size=(400, 400))
        bio = BytesIO()
        img.save(bio, format="png")

        self.set_header("Content-Type", "image/png")
        self.write(bio.getvalue())

    def get_mol(self):
        self.write(Chem.MolToMolBlock(self.mol))



import os
import signal
import socket
import argparse
import webbrowser
from contextlib import closing

import psutil
import tornado.web

from .db import connect
from .task_queue import TaskQueue
from .handler.app import AppInfoHandler
from .handler.calc import CalcIdHandler, CalcIdExtHandler
from .handler.file import FileHandler, FileIdHandler, FileIdExtHandler, FileIdNthExtHandler
from .handler.descriptor import DescriptorHandler
from .handler.singlefile import SingleFileHandler

MEGA = 1024 * 1024


class Shutdown(object):
    def __init__(self, server, ioloop):
        self.server = server
        self.ioloop = ioloop

    def __call__(self, sig, frame):
        print("catch signal: {}".format(sig))  # noqa: T003
        self.server.stop()
        self.ioloop.stop()


class MyApplication(tornado.web.Application):
    def __init__(self, queue, conn, file_size_limit, molecule_limit,
                 parse_timeout, prepare_timeout, calc_timeout, *args,
                 **kwargs):
        super(MyApplication, self).__init__(*args, **kwargs)
        self.queue = queue
        self.db = conn
        self.file_size_limit = file_size_limit
        self.parse_timeout = parse_timeout
        self.prepare_timeout = prepare_timeout
        self.molecule_limit = molecule_limit
        self.calc_timeout = calc_timeout


def get_free_address(lower=3000):
    port = lower
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        for i in range(5000):
            try:
                sock.bind(("", port + i))
            except OSError:
                pass
            else:
                return sock.getsockname()

        raise OSError("no free port")


def serve(port,
          workers,
          no_browser,
          file_size_limit=3,
          molecule_limit=50,
          parse_timeout=60,
          prepare_timeout=60,
          calc_timeout=60,
          db="mordred-web.sqlite"):

    if port is None:
        _, port = get_free_address()

    if workers is None:
        workers = psutil.cpu_count(logical=False)

    static = os.path.join(os.path.dirname(__file__), "static")
    ioloop = tornado.ioloop.IOLoop.current()

    with connect(db) as conn, TaskQueue(workers, ioloop) as queue:
        app = MyApplication(
            queue=queue,
            conn=conn,
            file_size_limit=file_size_limit,
            molecule_limit=molecule_limit,
            parse_timeout=parse_timeout,
            prepare_timeout=prepare_timeout,
            calc_timeout=calc_timeout,
            handlers=[
                (r"/api/descriptor", DescriptorHandler),
                (r"/api/info", AppInfoHandler),
                (r"/api/file", FileHandler),
                (r"/api/file/([0-9a-zA-Z]+)", FileIdHandler),
                (r"/api/file/([0-9a-zA-Z]+)\.(.*)", FileIdExtHandler),
                (r"/api/file/([0-9a-zA-Z]+)/([0-9]+)\.(.*)",
                 FileIdNthExtHandler),
                (r"/api/calc/([0-9a-zA-Z]+)", CalcIdHandler),
                (r"/api/calc/([0-9a-zA-Z]+)\.(.*)", CalcIdExtHandler),
                (r"/static/(.*)", tornado.web.StaticFileHandler, {
                    "path": static,
                }),
                (r"/.*", SingleFileHandler, {
                    "path": os.path.join(static, "index.html"),
                }),
            ],
            compress_response=True,
            static_hash_cache=True)
        server = tornado.httpserver.HTTPServer(
            app, max_body_size=(file_size_limit + 1) * MEGA)
        server.bind(port)
        server.start(1)
        url = "http://127.0.0.1:{}".format(port)
        if not no_browser:
            webbrowser.open(url, autoraise=True)

        signal.signal(signal.SIGTERM, Shutdown(server, ioloop))
        signal.signal(signal.SIGINT, Shutdown(server, ioloop))
        print("start mordred.web on {}".format(url))  # noqa: T003
        ioloop.start()
        print("terminate...")  # noqa: T003


def main():
    parser = argparse.ArgumentParser(description="Mordred Web UI")
    parser.add_argument("-p", "--port", type=int, default=None, help="port")
    parser.add_argument(
        "-w", "--workers", type=int, default=None, help="number of workers")
    parser.add_argument(
        "--file-size-limit",
        metavar="MB",
        type=int,
        default=5,
        help="upload file size limit", )
    parser.add_argument(
        "--molecule-limit",
        metavar="N",
        type=int,
        default=None,
        help="number of molecule limit", )
    parser.add_argument(
        "--parse-timeout",
        metavar="SEC",
        type=int,
        default=None,
        help="file parse timeout", )
    parser.add_argument(
        "--prepare-timeout",
        metavar="SEC",
        type=int,
        default=None,
        help="molecular preparation timeout", )
    parser.add_argument(
        "--calc-timeout",
        metavar="SEC",
        type=int,
        default=None,
        help="descriptor calculation timeout", )
    parser.add_argument(
        "--db",
        metavar="FILE",
        type=str,
        default="mordred-web.sqlite",
        help="database file path", )
    parser.add_argument(
        "--no-browser",
        action="store_true",
        default=False,
        help="don't open browser automatically", )
    result = (parser.parse_args())
    serve(**vars(result))


if __name__ == "__main__":
    main()



import json
from contextlib import contextmanager

from tornado import gen, web, iostream

from ..db import transaction


class RequestHandler(web.RequestHandler):
    @contextmanager
    def transaction(self):
        with transaction(self.db) as cur:
            yield cur

    def get_flag(self, name, default=object()):
        v = self.get_argument(name, default)

        if isinstance(v, bool):
            return v

        if v.lower() in {"false", "f"}:
            return False

        return True

    def put(self, task):
        return self.application.queue.put(task)

    @property
    def db(self):
        return self.application.db

    def fail(self, status, reason):
        raise web.HTTPError(status, reason=reason)

    def json(self, **kwargs):
        return self.write(kwargs)


class SSEHandler(RequestHandler):
    def init_sse(self):
        self.set_header("content-type", "text/event-stream")
        self.set_header("cache-control", "no-cache")

    @gen.coroutine
    def publish(self, **obj):
        self.write("data: {}\n\n".format(json.dumps(obj)))
        try:
            yield self.flush()
        except iostream.StreamClosedError:
            raise web.Finish



import math
import time
from cgi import parse_header

import openpyxl
from mordred import Calculator, descriptors
from tornado import gen, web
from mordred.error import MissingValueBase

from ..db import transaction, issue_text_id
from .common import SSEHandler, RequestHandler
from ..task_queue import Task, SingleTask


class PrepareTask(SingleTask):
    timeout = 60

    def __init__(self, calc_id, total, file_id, disabled, conn, calc_timeout):
        self.calc_id = calc_id
        self.file_id = file_id
        self.disabled = disabled
        self.conn = conn
        self.total = total
        self.calc_timeout = calc_timeout
        self.error = False

    def on_job_error(self, job, e):
        with transaction(self.conn) as cur:
            cur.execute(
                "INSERT INTO calc_error (calc_id, error) VALUES (?, ?)",
                (self.calc_id,
                 "BUG: calculator prepare failed: {!r}".format(e)), )
        self.error = True

    def on_job_end(self, job, calc):
        self.calc = calc
        self.desc_ids = []

        with transaction(self.conn) as cur:
            for desc in calc.descriptors:
                cur.execute(
                    "INSERT INTO descriptor (calc_id, name) VALUES (?, ?)",
                    (self.calc_id, str(desc)), )
                self.desc_ids.append(cur.lastrowid)

    def next_task(self):
        if self.error:
            return

        task = CalcTask(
            file_id=self.file_id,
            calc_id=self.calc_id,
            desc_ids=self.desc_ids,
            total=self.total,
            calc=self.calc,
            conn=self.conn,
            timeout=self.calc_timeout, )
        task.get_mols()
        return task

    def job(self):
        return PrepareWorker(disabled=self.disabled)


class PrepareWorker(object):
    def __init__(self, disabled):
        self.disabled = disabled

    def __call__(self):
        calc = Calculator(
            getattr(descriptors, d) for d in descriptors.__all__
            if d not in self.disabled)

        return calc


class CalcTask(Task):
    def __init__(self, file_id, calc_id, desc_ids, calc, total, conn, timeout):
        self.file_id = file_id
        self.calc_id = calc_id
        self.desc_ids = desc_ids
        self.calc = calc
        self.conn = conn
        self.total = total
        self.timeout = timeout

        Nd = len(desc_ids)
        self.max = [None] * Nd
        self.min = [None] * Nd
        self.mean = [None] * Nd
        self.M = [0.0] * Nd
        self.S = [0.0] * Nd
        self.k = [0] * Nd

    def get_mols(self):
        with transaction(self.conn) as cur:
            cur.execute(
                "SELECT id, mol FROM molecule WHERE file_id = ?",  # not require ORDER BY
                (self.file_id, ), )
            self.mols = cur.fetchall()

    def on_job_error(self, job, e):
        se = str(e)
        if len(se) == 0:
            se = repr(e)

        with transaction(self.conn) as cur:
            cur.execute(
                "INSERT INTO calc_error (calc_id, molecule_id, error) VALUES (?, ?, ?)",
                (self.calc_id, job.mol_id, se), )

    def on_task_end(self):
        std = ((None if k == 0 else math.sqrt(S / k))
               for S, k in zip(self.S, self.k))
        results = zip(self.desc_ids, self.min, self.max, self.mean, std)

        with transaction(self.conn) as cur:
            for desc_id, vmin, vmax, mean, std in results:
                cur.execute(
                    """
                    UPDATE descriptor
                    SET min = ?, max = ?, mean = ?, std = ?
                    WHERE id = ?""",
                    (vmin, vmax, mean, std, desc_id), )

            cur.execute("UPDATE calc SET done = 1 WHERE id = ?",
                        (self.calc_id, ))

    def on_job_end(self, job, results):

        with transaction(self.conn) as cur:
            for i, (desc_id, result) in enumerate(zip(self.desc_ids, results)):
                value, error = None, None
                if isinstance(result, MissingValueBase):
                    error = str(result.error)
                else:
                    value = result

                cur.execute("""
                    INSERT INTO result (calc_id, molecule_id, descriptor_id, value, error)
                    VALUES (?, ?, ?, ?, ?)
                    """, (self.calc_id, job.mol_id, desc_id, value, error))

                if error:
                    continue

                if self.max[i] is None or self.max[i] < value:
                    self.max[i] = value

                if self.min[i] is None or self.min[i] > value:
                    self.min[i] = value

                self.mean[i] = (self.mean[i] or 0) + value / self.total

                self.k[i] += 1

                M = self.M[i]
                self.M[i] += (value - M) / self.k[i]
                self.S[i] += (value - M) * (value - self.M[i])

            cur.execute(
                "UPDATE calc SET current = current + 1 WHERE id = ?",
                (self.calc_id, ), )

    def __next__(self):
        if len(self.mols) == 0:
            raise StopIteration
        (mol_id, mol), self.mols = self.mols[0], self.mols[1:]
        return CalcWorker(mol, mol_id, self.calc)


class CalcWorker(object):
    def __init__(self, mol, mol_id, calc):
        self.mol = mol
        self.mol_id = mol_id
        self.calc = calc

    def __call__(self):
        return self.calc(self.mol)


class CalcIdHandler(SSEHandler):
    def post(self, file_text_id):
        with self.transaction() as cur:
            cur.execute(
                "SELECT id, name, total FROM file WHERE text_id = ? LIMIT 1",
                (file_text_id, ), )
            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            file_id, file_name, total = result

            calc_text_id = issue_text_id()

            cur.execute("""
            INSERT INTO calc (file_id, text_id, created_at, current, done)
            VALUES (?, ?, ?, 0, 0)""", (file_id, calc_text_id,
                                        int(time.time())))

            calc_id = cur.lastrowid

        disabled = set(self.get_arguments("disabled"))
        task = PrepareTask(
            calc_id=calc_id,
            file_id=file_id,
            total=total,
            disabled=disabled,
            conn=self.db,
            calc_timeout=self.application.calc_timeout, )
        self.put(task)

        self.json(id=calc_text_id)

    def get(self, calc_text_id):
        with self.transaction() as cur:
            cur.execute(
                "SELECT id, file_id FROM calc WHERE text_id = ? LIMIT 1",
                (calc_text_id, ))
            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            self.calc_id, self.file_id = result

            cur.execute(
                "SELECT name, total, text_id FROM file WHERE id = ? LIMIT 1",
                (self.file_id, ), )
            self.file_name, self.total, file_text_id = cur.fetchone()

        accept, _ = parse_header(self.request.headers["Accept"])

        if accept == "text/event-stream":
            return self.get_sse()
        else:
            return self.get_json()

    @gen.coroutine
    def get_sse(self):
        self.init_sse()
        while True:
            with self.transaction() as cur:
                cur.execute(
                    "SELECT done, current FROM calc WHERE id = ? LIMIT 1",
                    (self.calc_id, ), )

                done, current = cur.fetchone()

            yield self.publish(
                total=self.total,
                name=self.file_name,
                done=bool(done),
                current=current, )

            if done:
                raise web.Finish
            yield gen.sleep(0.5)

    def get_json(self):
        with self.transaction() as cur:
            cur.execute(
                "SELECT name, text_id FROM file WHERE id = ? LIMIT 1",
                (self.file_id, ), )

            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            cur.execute("""
                SELECT molecule.nth, molecule.name, calc_error.error
                FROM calc_error LEFT OUTER JOIN molecule
                    ON calc_error.molecule_id = molecule.id
                WHERE calc_error.calc_id = ?
                ORDER BY molecule.nth
            """, (self.calc_id, ))

            errors = [{
                "error": error,
                "name": name,
                "nth": nth,
            } for nth, name, error in cur.fetchall()]

            cur.execute("""
                SELECT name, max, min, mean, std
                FROM descriptor
                WHERE calc_id = ?
                ORDER BY id
            """, (self.calc_id, ))

            descs = [{
                "max": vmax,
                "mean": mean,
                "min": vmin,
                "name": name,
                "std": std,
            } for name, vmax, vmin, mean, std in cur.fetchall()]

            file_name, file_text_id = result

            self.json(
                file_name=file_name,
                file_id=file_text_id,
                errors=errors,
                descriptors=descs, )


class CalcIdExtHandler(RequestHandler):
    EXTS = {"csv", "xlsx", "txt"}

    def get(self, calc_text_id, ext):
        ext = ext.lower()
        if ext not in self.EXTS:
            self.fail(400, "unknown extension")

        with self.transaction() as cur:
            cur.execute(
                "SELECT id, file_id FROM calc WHERE text_id = ? LIMIT 1",
                (calc_text_id, ))
            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            self.calc_id, self.file_id = result

            cur.execute(
                "SELECT id, name FROM molecule WHERE file_id = ? ORDER BY nth",
                (self.file_id, ), )
            self.molecules = cur.fetchall()

            cur.execute(
                "SELECT name FROM descriptor WHERE calc_id = ? ORDER BY id",
                (self.calc_id, ), )
            self.descriptors = [d for d, in cur.fetchall()]

            if ext == "csv":
                return self.get_csv(cur)
            elif ext == "xlsx":
                return self.get_xlsx(cur)
            elif ext == "txt":
                return self.get_error_log(cur)

    def get_error_log(self, cur):
        self.set_header("content-type", "text/plain")

        for mol_id, name in self.molecules:
            cur.execute("""
                SELECT error
                FROM calc_error
                WHERE calc_id = ? AND molecule_id = ?
                ORDER BY id
            """, (self.calc_id, mol_id))

            for e, in cur.fetchall():
                self.write("{}: {}\n".format(name, e))

            cur.execute("""
                SELECT descriptor.name, result.error
                FROM result JOIN descriptor ON result.descriptor_id = descriptor.id
                WHERE result.error IS NOT NULL AND result.calc_id = ? AND result.molecule_id = ?
                ORDER BY descriptor_id
            """, (self.calc_id, mol_id))

            for n, e in cur.fetchall():
                self.write('{}:{}: {}\n'.format(name, n, e))

    def _get_value_by_mol_id(self, cur, mol_id):
        cur.execute("""
            SELECT value
            FROM result
            WHERE calc_id = ? AND molecule_id = ?
            ORDER BY descriptor_id
            """, (self.calc_id, mol_id))
        return cur.fetchall()

    def get_csv(self, cur):
        self.set_header("content-type", "text/csv")
        self.write("name,")
        self.write(",".join(self.descriptors))
        self.write("\n")

        for mol_id, name in self.molecules:
            self.write(name + ",")
            result = self._get_value_by_mol_id(cur, mol_id)
            self.write(",".join(("" if v is None else str(v)) for v, in result))
            self.write("\n")

    def get_xlsx(self, cur):
        self.set_header(
            "content-type",
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(["name"] + self.descriptors)

        for mol_id, name in self.molecules:
            ws.append([name] + [v for v, in self._get_value_by_mol_id(cur, mol_id)])

        self.write(openpyxl.writer.excel.save_virtual_workbook(wb))



from setuptools import setup

setup(
    name='mordred-web',
    version='0.3.2',
    description='molecular descriptor calculator web UI',
    license='BSD3',
    author='Hirotomo Moriwaki',
    author_email='philopon.dependence@gmail.com',
    url='https://github.com/mordred-descriptor/mordred-web',
    platforms=['any'],
    keywords='QSAR chemoinformatics',
    packages=[
        'mordred.web',
        'mordred.web.handler',
    ],
    package_dir={
        'mordred.web': 'python/mordred_web',
        'mordred.web.handler': 'python/mordred_web/handler',
    },

    package_data={
        'mordred.web': [
            'static/*',
            'static/**/*',
        ],
    },

    install_requires=[
        'mordred>=0.3.1',
        'tornado>=4.5',
        'psutil>=5.0',
        'base58>=0.2.0',
        'numpy>=1.10',
        'six>=1.10',
        'loky>=1.0',
        'Pillow>=4.1'
        'openpyxl>=2.4',
    ],

)



import uuid
import sqlite3
from enum import Enum
from contextlib import closing, contextmanager

import base58
from rdkit import Chem


def issue_text_id():
    return base58.b58encode(uuid.uuid4().bytes)


@contextmanager
def transaction(conn):
    try:
        with closing(conn.cursor()) as cur:
            yield cur
    except Exception as e:
        conn.rollback()
        raise e
    else:
        conn.commit()


class Phase(Enum):
    PENDING = "pending"
    IN_PROGRESS = "in-progress"
    DONE = "done"
    ERROR = "error"


schema = [
    """
    CREATE TABLE IF NOT EXISTS file (
        id         INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        text_id    TEXT    NOT NULL,
        name       TEXT    NOT NULL,
        created_at INTEGER NOT NULL,
        desalt     INTEGER NOT NULL,
        gen3D      INTEGER NOT NULL,
        is3D       INTEGER NOT NULL,
        total      INTEGER,
        phase      TEXT    NOT NULL
    )
    """,
    """
    CREATE INDEX IF NOT EXISTS file__text_id ON file(text_id)
    """,
    """
    CREATE TABLE IF NOT EXISTS molecule (
        id         INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        file_id    INTEGER NOT NULL REFERENCES file(id) ON DELETE CASCADE ON UPDATE CASCADE,
        nth        INTEGER NOT NULL,
        forcefield TEXT,
        name       TEXT    NOT NULL,
        mol        MOL     NOT NULL,
        UNIQUE (file_id, nth)
    )
    """,
    """
    CREATE INDEX IF NOT EXISTS molecule__file_id ON molecule(file_id)
    """,
    """
    CREATE TABLE IF NOT EXISTS file_error (
        id      INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        file_id INTEGER NOT NULL REFERENCES file(id) ON DELETE CASCADE ON UPDATE CASCADE,
        error   TEXT    NOT NULL
    )
    """,
    """
    CREATE TABLE IF NOT EXISTS calc (
        id          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        file_id     INTEGER NOT NULL REFERENCES file(id) ON DELETE CASCADE ON UPDATE CASCADE,
        text_id     TEXT    NOT NULL,
        done        INTEGER NOT NULL,
        current     INTEGER NOT NULL,
        created_at  INTEGER NOT NULL
    )
    """,
    """
    CREATE INDEX IF NOT EXISTS calc__text_id ON calc(text_id)
    """,
    """
    CREATE TABLE IF NOT EXISTS descriptor (
        id      INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        calc_id INTEGER NOT NULL REFERENCES calc(id) ON DELETE CASCADE ON UPDATE CASCADE,
        name    TEXT NOT NULL,
        max     NUMBER,
        min     NUMBER,
        mean    NUMBER,
        std     NUMBER
    )
    """,
    """
    CREATE INDEX IF NOT EXISTS descriptor__calc_id ON descriptor(calc_id)
    """,
    """
    CREATE TABLE IF NOT EXISTS result (
        calc_id       INTEGER NOT NULL REFERENCES calc(id) ON DELETE CASCADE ON UPDATE CASCADE,
        molecule_id   INTEGER NOT NULL REFERENCES molecule(id) ON DELETE CASCADE ON UPDATE CASCADE,
        descriptor_id INTEGER NOT NULL REFERENCES descriptor(id) ON DELETE CASCADE ON UPDATE CASCADE,
        value   NUMBER,
        error   TEXT,
        UNIQUE (calc_id, molecule_id, descriptor_id)
    )
    """,  # noqa: E501
    """
    CREATE TABLE IF NOT EXISTS calc_error (
        id          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
        calc_id     INTEGER NOT NULL REFERENCES calc(id) ON DELETE CASCADE ON UPDATE CASCADE,
        molecule_id INTEGER REFERENCES molecule(id) ON DELETE CASCADE ON UPDATE CASCADE,
        error       TEXT NOT NULL
    )
    """,
]


def adapt_mol(m):
    return m.ToBinary()


def convert_mol(b):
    return Chem.Mol(b)


@contextmanager
def connect(db):
    sqlite_args = {
        "detect_types": sqlite3.PARSE_DECLTYPES,
        "isolation_level": "DEFERRED",
    }

    sqlite3.register_adapter(Chem.Mol, adapt_mol)
    sqlite3.register_converter("MOL", convert_mol)

    with sqlite3.connect(db, **sqlite_args) as conn:
        conn.text_factory = str
        conn.execute("PRAGMA foreign_keys = ON")
        with transaction(conn) as cur:
            for s in schema:
                cur.execute(s)

        yield conn



import os

from tornado.web import StaticFileHandler


class SingleFileHandler(StaticFileHandler):
    def initialize(self, path):
        dirname, self.filename = os.path.split(path)
        super(SingleFileHandler, self).initialize(dirname)

    def head(self, path=None):
        return super(SingleFileHandler, self).head(self.filename)

    def get(self, path=None, include_body=True):
        return super(SingleFileHandler, self).get(self.filename, include_body=include_body)



from mordred import descriptors

from .common import RequestHandler


class DescriptorHandler(RequestHandler):
    def get(self):
        self.write({
            "descriptors": descriptors.__all__,
        })
